#ifndef _Poly_H
#define _Poly_H

/*
******************************************************************************************
	Main Header for Poly
	NOTE:
		Look carefully at my code to check for poly/plane classification
		It is very different than others
			In every other code that iv seen, given a quad defined as points (p0,p1,p2,p3),
				|
				|
			p0*-------* p1
				|				|
				|				|
			p3*-------* p2
				|
				|
			if p0 & p3 were CO_PLANAR and 
				 p1 & p2 were FRONT
			the poly would be split     HELL NO!!!!!
			it could be considered as FRONT in this case since no front->back crossover exists
			ALL other code i have come accros assumes that just because the classification changes
			(ie from CO_PLANAR to FRONT) it must be split
		More robust this way
******************************************************************************************
*/

class Poly;
typedef list<Poly>  	Poly_List;
Poly_List::iterator		poly_Iter;

void SCR_project();
void SCR_render();
bool clip(int);

extern int just_traverse_BSP, no_offscreen_draw;

class Plane {
public:
	Vect3D normal;
	double dist_origin;

	Plane(): dist_origin(0) {}
	Plane(const Plane &plane):normal(plane.normal), dist_origin(plane.dist_origin) {}

	Vect3D point_on_plane() const; // find a point on the plane, trying not to DIV_0
	double dist          (const Point3D &) const; // spatial determination [front/back test]
	int    classify      (const Poly &) const;

	Plane &operator=(const Plane &p);
 };
//******************************************************************************************

class Poly {
protected:
	void transform();
	bool clip_to_fustrum();
public:
	static Poly clipped_poly;

	Vect3D vertex[MAX_VERTEXES];
	Plane  plane;
	int num_vertexes;
	bool used;

	Poly() :num_vertexes(0), used(false) {}
	Poly(const Poly &);
	~Poly() {}

	bool operator==(const Poly &);
	Poly& operator=(const Poly &);
	void set(const Poly &);					// copy only num_vertexes and vertex array    used in clip

	void clear() {num_vertexes=0;}
	void add(const Point3D &);
	void us_add(const Point3D &);  // UnSafe add   does not accound for current vertex number
	void solve();
	void draw();
};
//******************************************************************************************

Plane &Plane::operator=(const Plane &rhs) {
	normal=rhs.normal;
	dist_origin=rhs.dist_origin;
	return *this;
}

int  Plane::classify(const Poly &poly) const {
	double d;
	register int i;
	bool front, back, coplanar;

	front=back=coplanar=false;

	if((poly.plane.normal==normal) && (poly.plane.dist_origin==dist_origin)) {
		return CO_PLANAR;
	}

	for(i=0;i<poly.num_vertexes; i++) {
		d=dist(poly.vertex[i]);

		if(d > -CLIP_PLANE_EPSILON  &&  d < CLIP_PLANE_EPSILON) {
			coplanar=true;
		}
		if(d>CLIP_PLANE_EPSILON) {
			front=true;
		}
		if(d<-CLIP_PLANE_EPSILON) {
			back=true;
		}
	}
	if(front && back) return SPANNING;
	if(coplanar && front) return FRONT;
	if(coplanar && back)  return BACK;
	if(front) 	 return FRONT;
	if(back) 		 return BACK;
	return CO_PLANAR;
}

Vect3D Plane::point_on_plane() const{
	if(normal.getx())
		return Vect3D(dist_origin/normal.getx(),0,0);
	else if(normal.gety())
		return Vect3D(0, dist_origin/normal.gety(),0);
	else
		return Vect3D(0,0,dist_origin/normal.getz());  // FIXME: Possible DIV_0
}

double Plane::dist(const Point3D &point_in_space) const {
	return (point_in_space-point_on_plane()).dot(normal);
}
//******************************************************************************************
Poly::Poly(const Poly &poly): plane(poly.plane), num_vertexes(poly.num_vertexes), used(poly.used) {
	memcpy(vertex, poly.vertex, num_vertexes*sizeof(Vect3D));
}

bool Poly::operator==(const Poly &p) {
	if(num_vertexes!=p.num_vertexes) return false;
	for(int i=0; i<num_vertexes; i++)
		if(vertex[i]!=p.vertex[i]) return false;
	return true;
}

Poly& Poly::operator=(const Poly &rhs) {
	plane=rhs.plane;
	used=rhs.used;
	num_vertexes=rhs.num_vertexes;
	memcpy(vertex, rhs.vertex, num_vertexes*sizeof(Vect3D));
	return *this;
}
void Poly::set(const Poly &rhs) {
	num_vertexes=rhs.num_vertexes;
	memcpy(vertex, rhs.vertex, num_vertexes*sizeof(Vect3D));
}	

void Poly::solve() {
	Vect3D u,v;

	u=vertex[1]-vertex[0];
	v=vertex[2]-vertex[0];

	plane.normal=u.cross(v);	// find raw normal
	plane.normal.insure_0();  // make sure its not something like   -0.0
	plane.normal.norm();			// normalize
	plane.dist_origin=plane.normal.dot(vertex[0]);
}

void Poly::add(const Point3D &v) {
	for(int i=0; i<num_vertexes; i++)  // must check for duplicated points because of way my clip works
		if(vertex[i]==v) return;

	vertex[num_vertexes++]=v;
}
void Poly::us_add(const Point3D &v) {
	vertex[num_vertexes++]=v;
}

void Poly::transform() {
	clipped_poly.clear();
	for(int i=0; i<num_vertexes; i++)
		clipped_poly.us_add(player.orientation.cross(vertex[i]-player.pos));

}
bool Poly::clip_to_fustrum() {
	if(!clip(LEFT)) return false;
	if(!clip(RIGHT)) return false;
	if(!clip(BOTTOM)) return false;
	return clip(TOP);
}

void Poly::draw() {
	if(just_traverse_BSP) return;

	transform();  // object -> world
	if(!clip_to_fustrum())
		return;

	if(no_offscreen_draw) return;
	SCR_project(); // world -> screen
	SCR_render();		// screen ->  well.... to the lowlevel rasterizer for the  SCREEN  doh!
}
//******************************************************************************************
Poly Poly::clipped_poly;
#endif